package cn.dankal.test.first;

import cn.dankal.test.common.Const;
import cn.hutool.core.util.RandomUtil;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.StringSerializer;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.ExecutionException;

/**
 * 测试 Kafka clients: 第一个 producer
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-25
 */
public class KafkaFirstProducer {


    /**
     * KafkaProducer
     */
    private static KafkaProducer<String, String> kafkaProducer;


    @BeforeAll
    public static void before() {
        Properties properties = new Properties();
        final String bootstrapServers = "192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094";
        // broker(Kafka服务器) 地址清单
        properties.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers);
        // message key 序列化器
        properties.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getCanonicalName());
        // message value 序列化器
        properties.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getCanonicalName());

        // 通过 Properties 配置 实例化 KafkaProducer
        kafkaProducer = new KafkaProducer<>(properties);
    }

    /**
     * 第一个 Kafka 生产者
     */
    @Test
    public void testKafkaFirstProducer() throws ExecutionException, InterruptedException {
        final int sendMessageCount = 6;
        for (int i = 0; i < sendMessageCount; i++) {
            // 消息key (可以为空)
            final String key = "dankal-hello-" + RandomUtil.randomInt(10, 5000);
            // 消息内容
            final String message = RandomUtil.randomString(64);

            ProducerRecord<String, String> stringProducerRecord =
                    new ProducerRecord<String, String>(
                            Const.Kafka.Topic.HELLO_TOPIC,
                            key,
                            message
                    );
            kafkaProducer.send(stringProducerRecord).get();
        }
    }

    @AfterAll
    public static void after() {
        if (Objects.nonNull(kafkaProducer)) {
            kafkaProducer.close();
        }
    }
}
