package cn.dankal.test.boot.config;

import cn.dankal.test.boot.common.Const;
import org.apache.kafka.clients.admin.AdminClientConfig;
import org.apache.kafka.clients.admin.NewTopic;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.kafka.KafkaProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.kafka.core.KafkaAdmin;
import org.springframework.util.StringUtils;

import java.util.HashMap;
import java.util.Map;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-22
 */
@Configuration
public class KafkaTopicConfig {

    /**
     * KafkaProperties
     */
    private final KafkaProperties kafkaProperties;

    /**
     * Kafka 集群数量
     */
    @Value("${spring.kafka.broker-count}")
    private Short brokerCount;

    public KafkaTopicConfig(KafkaProperties kafkaProperties) {
        this.kafkaProperties = kafkaProperties;
    }

    @Bean
    public KafkaAdmin admin() {
        Map<String, Object> configs = new HashMap<>(1 << 5);
        configs.put(AdminClientConfig.BOOTSTRAP_SERVERS_CONFIG,
                StringUtils.arrayToCommaDelimitedString(kafkaProperties.getBootstrapServers().toArray()));
        return new KafkaAdmin(configs);
    }

    /**
     * 新建一个kafka主题: hello-spring-kafka 如果需要话
     *
     * @return NewTopic
     * @see Const.Kafka.Topic#TOPIC_HELLO_SPRING_KAFKA
     */
    @Bean
    public NewTopic helloSpringKafkaTopic() {
        return new NewTopic(
                Const.Kafka.Topic.TOPIC_HELLO_SPRING_KAFKA,
                kafkaProperties.getListener().getConcurrency(),
                brokerCount);
    }

    /**
     * 新建一个kafka主题: simple-string-message 如果需要话
     *
     * @return NewTopic
     * @see Const.Kafka.Topic#TOPIC_SIMPLE_STRING_MESSAGE
     */
    @Bean
    public NewTopic simpleStringKafkaTopic() {
        return new NewTopic(
                Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE,
                kafkaProperties.getListener().getConcurrency(),
                brokerCount);
    }

    /**
     * 新建一个kafka主题: handler-string-message-error-message 如果需要话
     *
     * @return NewTopic
     * @see Const.Kafka.Topic#TOPIC_HANDLER_STRING_MESSAGE_ERROR_MESSAGE
     */
    @Bean
    public NewTopic handlerStringMessageErrorMessageKafkaTopic() {
        return new NewTopic(
                Const.Kafka.Topic.TOPIC_HANDLER_STRING_MESSAGE_ERROR_MESSAGE,
                kafkaProperties.getListener().getConcurrency(),
                brokerCount);
    }


}
