package cn.dankal.test.boot.handler;

import cn.dankal.test.boot.common.Common;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.consumer.Consumer;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.kafka.listener.ConsumerAwareListenerErrorHandler;
import org.springframework.kafka.listener.ListenerExecutionFailedException;
import org.springframework.kafka.support.KafkaHeaders;
import org.springframework.messaging.Message;
import org.springframework.messaging.MessageHeaders;
import org.springframework.messaging.handler.annotation.SendTo;
import org.springframework.stereotype.Component;

/**
 * {@link ConsumerAwareListenerErrorHandler}
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-29
 * @see SendTo
 */
@Slf4j
@Component
@ConditionalOnBean(name = {"testErrorHandler"})
public class KafkaStringMessageListenerErrorHandler implements ConsumerAwareListenerErrorHandler {

    @Override
    public Object handleError(Message<?> message, ListenerExecutionFailedException exception, Consumer<?, ?> consumer) {
        Throwable exceptionCause = exception.getCause();
        String errorMessage = exception.getMessage();
        if (null != exceptionCause) {
            errorMessage = exceptionCause.getMessage();
        }

        // TODO 消息信息获取
        // TODO org.springframework.kafka.support.KafkaHeaders
        MessageHeaders messageHeaders = message.getHeaders();
        // doc:  The header containing the topic from which the message was received.
        String receivedTopic = messageHeaders.get(KafkaHeaders.RECEIVED_TOPIC, String.class);
        Integer partitionId = messageHeaders.get(KafkaHeaders.RECEIVED_PARTITION_ID, Integer.class);
        Long offset = messageHeaders.get(KafkaHeaders.OFFSET, Long.class);
        if (log.isDebugEnabled()) {
            log.debug(
                    "线程 id: {} receivedTopic: {} partitionId: {} offset: {}",
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    receivedTopic,
                    partitionId,
                    offset
            );
        }
        // TODO  处理异常消息,做一些补偿操作
        // TODO 诸如 使用 Consumer#seek(TopicPartition partition, long offset) 重置 偏移量
        // TODO 让消费者重新消费
        if (log.isInfoEnabled()) {
            log.info(
                    "{} 处理字符串消息: {} 出现异常:{} ",
                    Thread.currentThread().getName(),
                    message.getPayload(),
                    errorMessage);
        }
        // TODO 因为测试用例 使用 JsonSerializer 处理的数据 无法实现 使用 String 消费
        // 异常摘要: Cannot convert from [cn.dankal.test.boot.entity.DankalSimpleMessage] to [java.lang.String]
        // 故在此出选择提交当前消息的偏移量 提交避免重复通知
        consumer.commitSync();
        // @SendTo
        // 此方法返回将会被 SendTo 注解作为消息发送到指定主题
        return "出现异常啦: " + errorMessage + "  !!!";
    }
}
