package cn.dankal.test.boot.listener;

import cn.dankal.test.boot.common.Common;
import cn.dankal.test.boot.common.Const;
import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.autoconfigure.condition.ConditionalOnMissingBean;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.support.Acknowledgment;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-28
 */
@Slf4j
@Component
@ConditionalOnMissingBean(name = {"testErrorHandler","test-retry"})
@ConditionalOnProperty(prefix = "spring.kafka",
        name = {"listener.type"},havingValue = "single")
public class StringKafkaMessageListener {

    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_SIMPLE_STRING_CONSUMER
    )
    public void stringMessageListener(String message,
                                      Acknowledgment acknowledgment) {
        if (log.isInfoEnabled()) {
            log.info("{} 执行线程: {} 消费消息 ==> 主题: {}",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    message);
        }
        acknowledgment.acknowledge();
    }
}
