# 蛋壳创意科技-技术测试-Kafka集成SpringBoot

> company: [蛋壳创意科技](https://www.dankal.cn/)
>
> code: [dankal-test-kafka](http://git.dankal.cn/mercyModest/dankal-test-kafka.git)
>
> spring-kafka-doc:  
>
> - **html :  [spring-kafka-2.2.8.RELEASE-html](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/html/)**
> - **pdf:  [spring-kafka-2.2.8.RELEASE-pdf](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/pdf/index.pdf)**

## 测试用例依赖版本

> - ==spring-boot: 	2.1.8.RELEASE==
>   - spring-kafka:	2.2.8.RELEASE
>     - kafka-clients:	2.0.1
>
> 其实 `spring-kafka`和`kafka-clients`的版本被统一管理了,我们只需要关注 spring boot 版本即可
>
> - jdk:	1.8.0_241
> - gradle:   Gradle 6.7.1
>
> ### Kafka 服务相关
>
> - kafka server:    2.7.0  [集群：3]
> - zookeeper:       2.4.0  [集群:   3]

## spring-kafka: Hello World

具体可以参阅: [dankal-test-kafka-build.gradle](https://git.dankal.cn/mercyModest/dankal-test-kafka/blob/master/build.gradle)

### 添加依赖

```groovy
dependencies {
    // spring dependence
    implementation 'org.springframework.boot:spring-boot-starter-web'
    implementation 'org.springframework.kafka:spring-kafka'
    testImplementation('org.springframework.boot:spring-boot-starter-test') {
        exclude group: 'org.junit.vintage', module: 'junit-vintage-engine'
    }
}
```

### 主配置文件: `application-native.yml`

```yml
server:
  port: 9001

spring:
  kafka:
    # kafka集群数量
    # 用于配置 replicationFactor : 显示创建 topic 时,设置的 partition 副本基数
    broker-count: 3
    # Kafka 集群信息
    bootstrap-servers:
      - 192.168.128.3:9092
    #      - 192.168.128.3:9093
    #      - 192.168.128.3:9094
    listener:
      # 消费者确认模式: 手动
      ack-mode: manual_immediate
      # 开启多少个消费者线程
      concurrency: 3
      # 一次确认消费的数量
      ack-count: 1
      type: single
      ack-time: 60
    template:
      # 默认 topic 名称
      default-topic: dankal-test-default-topic
    consumer:
      # 采用手动提交
      enable-auto-commit: false
      # 消息使用 Spring 提供的 JSON 发序列化器
      value-deserializer: org.springframework.kafka.support.serializer.JsonDeserializer
      # 设置为最初的消费进度
      # 其取值具体区别可以参阅
      # https://blog.csdn.net/lishuangzhe7047/article/details/74530417
      auto-offset-reset: earliest
      properties:
        # 默认信任 java.util,java.lang 包下面的类
        # 因此我们需要添加 spring json trusted 配置
        spring:
          json:
            trusted:
              packages: cn.dankal.test.boot.entity
    producer:
      # 发送失败,重试次数
      retries: 5
      # 只需要 leader 应答即可
      acks: 1
      # 消息内容使用 Spring 提供的 JSON 序列化器
      value-serializer: org.springframework.kafka.support.serializer.JsonSerializer



logging:
  level:
    cn:
      dankal:
        test: debug
    # 日志太多了,方便观察测试日志 --> 先调高日志级别  ==> error
    org:
      apache:
        kafka: error
      springframework:
        kafka: error
```

### 使用配置类创建需要的 topic

```java
package cn.dankal.test.boot.config;

import cn.dankal.test.boot.common.Const;
import org.apache.kafka.clients.admin.AdminClientConfig;
import org.apache.kafka.clients.admin.NewTopic;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.kafka.KafkaProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.kafka.core.KafkaAdmin;
import org.springframework.util.StringUtils;

import java.util.HashMap;
import java.util.Map;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-22
 */
@Configuration
public class KafkaTopicConfig {

    /**
     * KafkaProperties
     */
    private final KafkaProperties kafkaProperties;

    /**
     * Kafka 集群数量
     */
    @Value("${spring.kafka.broker-count}")
    private Short brokerCount;

    public KafkaTopicConfig(KafkaProperties kafkaProperties) {
        this.kafkaProperties = kafkaProperties;
    }

    @Bean
    public KafkaAdmin admin() {
        Map<String, Object> configs = new HashMap<>(1 << 5);
        configs.put(AdminClientConfig.BOOTSTRAP_SERVERS_CONFIG,
                StringUtils.arrayToCommaDelimitedString(kafkaProperties.getBootstrapServers().toArray()));
        return new KafkaAdmin(configs);
    }

    /**
     * 新建一个kafka主题: hello-spring-kafka 如果需要话
     *
     * @return NewTopic
     */
    @Bean
    public NewTopic helloSpringKafkaTopic() {
        return new NewTopic(
                Const.Kafka.Topic.TOPIC_HELLO_SPRING_KAFKA,
                kafkaProperties.getListener().getConcurrency(),
                brokerCount);
    }

    /**
     * 新建一个kafka主题: simple-string-message 如果需要话
     *
     * @return NewTopic
     */
    @Bean
    public NewTopic simpleStringKafkaTopic() {
        return new NewTopic(
                Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE,
                kafkaProperties.getListener().getConcurrency(),
                brokerCount);
    }
}
```

### 消息传递实体

```java
package cn.dankal.test.boot.entity;

import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import com.fasterxml.jackson.datatype.jsr310.deser.LocalDateTimeDeserializer;
import com.fasterxml.jackson.datatype.jsr310.ser.LocalDateTimeSerializer;
import lombok.Data;
import lombok.experimental.Accessors;
import org.springframework.kafka.support.serializer.JsonSerializer;

import java.io.Serializable;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.Date;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-28
 */
@Data
@Accessors(chain = true)
public class DankalSimpleMessage implements Serializable {
    private static final long serialVersionUID = -81606093254258919L;

    /**
     * message is
     */
    private Integer id;

    /**
     * message key
     */
    private String key;

    /**
     * message value
     */
    private String content;

    /**
     * message send time
     * <pre>
     * 因为  {@link JsonSerializer} 默认使用支持 {@link Date} 的序列化和反序列化对于Java8提供的诸如
     * {@link LocalDateTime}
     * {@link LocalDate}
     * {@link LocalTime}
     * 新的时间API需要通过注解指定相关序列化和反序列化器
     * </pre>
     */
    @JsonSerialize(using = LocalDateTimeSerializer.class)
    @JsonDeserialize(using = LocalDateTimeDeserializer.class)
    private LocalDateTime sendTime;
}
```

### 使用SpringKafka通过三种形式发送消息

#### **类的头部**

```java
  /**
     * KafkaTemplate<String, DankalSimpleMessage>
     */
    private final KafkaTemplate<String, DankalSimpleMessage> kafkaTemplate;

    /**
     * DankalMessageSuccessCallback
     */
    private final DankalMessageSuccessCallback dankalMessageSuccessCallback;

    /**
     * DankalMessageFailureCallback
     */
    private final AsyncSendMessageFailureCallback asyncSendMessageFailureCallback;

    public KafkaDankalMessageServiceImpl(KafkaTemplate<String, DankalSimpleMessage> kafkaTemplate, DankalMessageSuccessCallback dankalMessageSuccessCallback, AsyncSendMessageFailureCallback asyncSendMessageFailureCallback) {
        this.kafkaTemplate = kafkaTemplate;
        this.dankalMessageSuccessCallback = dankalMessageSuccessCallback;
        this.asyncSendMessageFailureCallback = asyncSendMessageFailureCallback;
    }
```
#### **发送并忘记**

  ```java
  @Override
  public boolean sendMessageBySimple(String topic, String key, DankalSimpleMessage dankalSimpleMessage) {
          // 发送并忘记
          kafkaTemplate.send(topic, key, dankalSimpleMessage);
          return true;
  }
  ```

#### **同步发送**

```java
@Override
public boolean sendMessageBySync(String topic, String key, DankalSimpleMessage dankalSimpleMessage) throws ExecutionException, InterruptedException {
        // 同步发送
        SendResult<String, DankalSimpleMessage> sendResult = kafkaTemplate.send(topic, key, dankalSimpleMessage).get();
        return sendResult.getRecordMetadata().offset() > 0;
}
```

#### **异步发送**

```java
@Override
public boolean sendMessageByAsync(String topic, String key, DankalSimpleMessage dankalSimpleMessage) {
        // 异步发送
        kafkaTemplate
                .send(topic, key, dankalSimpleMessage)
                // 添加异步回调处理
                .addCallback(dankalMessageSuccessCallback, asyncSendMessageFailureCallback);
        return true;
}
```

**对于异步发送，我们可以设置回调函数**

##### 举个栗子

###### 成功回调

```java
package cn.dankal.test.boot.component;

import cn.dankal.test.boot.entity.DankalSimpleMessage;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.springframework.kafka.support.SendResult;
import org.springframework.stereotype.Component;
import org.springframework.util.concurrent.SuccessCallback;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-28
 */
@Slf4j
@Component
public class DankalMessageSuccessCallback implements SuccessCallback<SendResult<String, DankalSimpleMessage>> {

    @Override
    public void onSuccess(SendResult<String, DankalSimpleMessage> result) {
        if (null == result) {
            if (log.isWarnEnabled()) {
                log.warn("current result is null!!!");
            }
            return;
        }
        RecordMetadata recordMetadata = result.getRecordMetadata();
        if (log.isDebugEnabled() && null != recordMetadata) {
            log.debug(
                    "成功发送消息：主题: {}, 分区: {},偏移量: {} ,key: {}, value: {}",
                    recordMetadata.topic(),
                    recordMetadata.partition(),
                    recordMetadata.offset(),
                    result.getProducerRecord().key(),
                    result.getProducerRecord().value()
            );
        }
    }
}
```

###### 失败回调

```java
package cn.dankal.test.boot.component;

import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;
import org.springframework.util.concurrent.FailureCallback;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-28
 */
@Slf4j
@Component
public class AsyncSendMessageFailureCallback implements FailureCallback {

    @Override
    public void onFailure(Throwable ex) {
        Throwable exCause = ex.getCause();
        log.error("异步:发送消息失败 ==>异常原因:{} 异常消息: {}", exCause != null ? exCause.getMessage() : "", ex.getMessage());
    }
}
```

### SpringKafka对消息的监听(消费)

> 具体的监听方法的参数形式，可以参阅文档
>
> [receiving-messages](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/html/#receiving-messages)
>
> 传输截图
>
> ![](https://img.mercymodest.com/public/20210129100450.png)
>
> ![](https://img.mercymodest.com/public/20210129100504.png)

#### 举个栗子

```java
package cn.dankal.test.boot.listener;

import cn.dankal.test.boot.common.Common;
import cn.dankal.test.boot.common.Const;
import cn.dankal.test.boot.entity.DankalSimpleMessage;
import lombok.extern.slf4j.Slf4j;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.support.Acknowledgment;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-21
 */
@Slf4j
@Component
public class KafkaDankalMessageListener {


    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_HELLO_SPRING_KAFKA},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_HELLO_SPRING_KAFKA_CONSUMER
    )
    public void dankalMessageListener(DankalSimpleMessage dankalSimpleMessage,
                                      Acknowledgment acknowledgment) {
        if (log.isInfoEnabled()) {
            log.info("{} 执行线程: {} 消费消息 ==> 主题: {}",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    dankalSimpleMessage);
        }
        acknowledgment.acknowledge();
    }
}
```

### 小小节

以上呢就是Spring-Kafka 的基本使用示例啦，具体可以参阅官方文档

**传送门:	[Using Spring for Apache Kafka](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/html/#kafka)**

## spring-kafka: 消息异常处理机制示例

### 测试监听器

> ```java
> @KafkaListener(
>       ... ...
>             // 设置 errorHandler 的 spring bean 名称
>             errorHandler = "kafkaStringMessageListenerErrorHandler"
>     )
> // 将 errorHandler 的处理方法返回值发送到那个主题
> @SendTo({Const.Kafka.Topic.TOPIC_HANDLER_STRING_MESSAGE_ERROR_MESSAGE})
> public String testStringMessageHandlerError(String message, Acknowledgment acknowledgment)
> ```

```java
package cn.dankal.test.boot.listener;

import cn.dankal.test.boot.common.Common;
import cn.dankal.test.boot.common.Const;
import cn.dankal.test.boot.handler.KafkaStringMessageListenerErrorHandler;
import lombok.extern.slf4j.Slf4j;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.support.Acknowledgment;
import org.springframework.messaging.handler.annotation.SendTo;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-29
 */
@Slf4j
@Component("testErrorHandler")
public class KafkaListenerErrorHandlerTestListener {

    /**
     * <pre>
     *     此监听器用来测试 spring kafka 对 消息消费异常的处理机制
     *     此监听器接收的 String 类型的 message,但是其订阅 主题 所传递的消息是: DankalMessage
     *     故将会抛出异常,以达到测试 errorHandler 的作用
     * </pre>
     *
     * @param message        需要消息消息
     * @param acknowledgment 消息确认器
     * @return 异常消息
     */
    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_HELLO_SPRING_KAFKA},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_TEST_STRING_MESSAGE_HANDLER_ERROR_CONSUMER,
            errorHandler = "kafkaStringMessageListenerErrorHandler"
    )
    @SendTo({Const.Kafka.Topic.TOPIC_HANDLER_STRING_MESSAGE_ERROR_MESSAGE})
    public String testStringMessageHandlerError(String message, Acknowledgment acknowledgment) {
        if (log.isInfoEnabled()) {
            log.info("{} 执行线程: {} 消费消息 ==> 主题: {}",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    message);
        }
        acknowledgment.acknowledge();
        return "handler message: " + message + " success";
    }

    /**
     * <pre>
     * 此监听器 是用来处理 {@link #testStringMessageHandlerError(String, Acknowledgment)} 消费消息时出现异常时,
     * {@link KafkaStringMessageListenerErrorHandler} 处理方法的方法返回值.
     * </pre>
     *
     * @param message        异常处理方法返回值
     * @param acknowledgment 消息确认器
     */
    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_HANDLER_STRING_MESSAGE_ERROR_MESSAGE},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_TEST_STRING_MESSAGE_HANDLER_ERROR_CONSUMER
    )
    public void handlerStringMessageError(String message, Acknowledgment acknowledgment) {
        if (log.isInfoEnabled()) {
            log.info("handlerStringMessageError ==> {} 执行线程: {} 处理异常信息 ==> 异常消息: {}",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    message);
        }
        acknowledgment.acknowledge();
    }
}
```

### 测试异常处理器

```java
package cn.dankal.test.boot.handler;

import cn.dankal.test.boot.common.Common;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.consumer.Consumer;
import org.springframework.kafka.listener.ConsumerAwareListenerErrorHandler;
import org.springframework.kafka.listener.ListenerExecutionFailedException;
import org.springframework.kafka.support.KafkaHeaders;
import org.springframework.messaging.Message;
import org.springframework.messaging.MessageHeaders;
import org.springframework.messaging.handler.annotation.SendTo;
import org.springframework.stereotype.Component;

/**
 * {@link ConsumerAwareListenerErrorHandler}
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-29
 * @see SendTo
 */
@Slf4j
@Component
public class KafkaStringMessageListenerErrorHandler implements ConsumerAwareListenerErrorHandler {

    @Override
    public Object handleError(Message<?> message, ListenerExecutionFailedException exception, Consumer<?, ?> consumer) {
        Throwable exceptionCause = exception.getCause();
        String errorMessage = exception.getMessage();
        if (null != exceptionCause) {
            errorMessage = exceptionCause.getMessage();
        }

        // TODO 消息信息获取
        // TODO org.springframework.kafka.support.KafkaHeaders
        MessageHeaders messageHeaders = message.getHeaders();
        // doc:  The header containing the topic from which the message was received.
        String receivedTopic = messageHeaders.get(KafkaHeaders.RECEIVED_TOPIC, String.class);
        Integer partitionId = messageHeaders.get(KafkaHeaders.RECEIVED_PARTITION_ID, Integer.class);
        Long offset = messageHeaders.get(KafkaHeaders.OFFSET, Long.class);
        if (log.isDebugEnabled()) {
            log.debug(
                    "线程 id: {} receivedTopic: {} partitionId: {} offset: {}",
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    receivedTopic,
                    partitionId,
                    offset
            );
        }
        // TODO  处理异常消息,做一些补偿操作
        // TODO 诸如 使用 Consumer#seek(TopicPartition partition, long offset) 重置 偏移量
        // TODO 让消费者重新消费
        if (log.isInfoEnabled()) {
            log.info(
                    "{} 处理字符串消息: {} 出现异常:{} ",
                    Thread.currentThread().getName(),
                    message.getPayload(),
                    errorMessage);
        }
        // TODO 因为测试用例 使用 JsonSerializer 处理的数据 无法实现 使用 String 消费
        // 异常摘要: Cannot convert from [cn.dankal.test.boot.entity.DankalSimpleMessage] to [java.lang.String]
        // 故在此出选择提交当前消息的偏移量 提交避免重复通知
        consumer.commitSync();
        // @SendTo
        // 此方法返回将会被 SendTo 注解作为消息发送到指定主题
        return "出现异常啦: " + errorMessage + "  !!!";
    }
}
```

### 小说明

#### 关于 spring-kafka 的 异常处理

> **详情可以参阅官方文档: [Handling Exceptions](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/html/#annotation-error-handling)**

#### 关于 spring-kafka 的 @SendTo 

> **详情可以参阅官方文档:[Forwarding Listener Results using `@SendTo`](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/html/#annotation-send-to)**

## spring-kafka: 批量发送消息

### spring-kafka 批量发送配置

> - 【数量】`batch-size` ：超过收集的消息数量的最大条数
>   - `spring.kafka.producer.batch-size`
> - 【空间】`buffer-memory` ：超过收集的消息占用的最大内存
>   - `spring.kafka.producer.buffer-memory`
> - 【时间】`linger.ms` ：超过收集的时间的最大等待时长，单位：毫秒
>   - `spring.kafka.producer.properties.linger.ms`

### etc.

**为了测试方便，测试用例使用的 `linger.ms`来测试 消息的批量发送的，设置的值为: ==30s==**

```yaml
server:
  port: 9001

... ...

spring:
  kafka:
  ... ...
    producer:
  ... ...
      ## 以下配置用于测试 spring-kafka 的批量发送消息
      batch-size: 16384
      buffer-memory: 33554432
      properties:
      # 超过收集的时间的最大等待时长
          linger:
             ms: 30000
... ...
```

### controller

```java
    @GetMapping("/test/string")
    public String testSendStringMessage() {
        final String topic = Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE;
        final String key = "dankal-test-str-" + RandomUtil.randomInt(10, 500);
        final String message = RandomUtil.randomString(20);
        kafkaStringMessageService.sendStringMessage(topic, key, message);
        return "success";
    }
```

### service

```java
    @Override
    public boolean sendStringMessage(String topic, String key, String message) {
        stringKafkaTemplate.send(topic, key, message);
        return true;
    }
```

### listener

```java
    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_SIMPLE_STRING_CONSUMER
    )
    public void stringMessageListener(String message,
                                      Acknowledgment acknowledgment) {
        if (log.isInfoEnabled()) {
            log.info("{} 执行线程: {} 消费消息 ==> 主题: {}",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    message);
        }
        acknowledgment.acknowledge();
    }
```

### 执行测试

浏览器==间隔==发送三次请求

> http://localhost:9001/test/string
>
> http://localhost:9001/test/string
>
> http://localhost:9001/test/string

### console 控制台打印

```shell
2021-01-29 16:01:25.099  INFO 11620 --- [nio-9001-exec-1] o.a.c.c.C.[Tomcat].[localhost].[/]       : Initializing Spring DispatcherServlet 'dispatcherServlet'
2021-01-29 16:01:25.099  INFO 11620 --- [nio-9001-exec-1] o.s.web.servlet.DispatcherServlet        : Initializing Servlet 'dispatcherServlet'
2021-01-29 16:01:25.103  INFO 11620 --- [nio-9001-exec-1] o.s.web.servlet.DispatcherServlet        : Completed initialization in 4 ms

2021-01-29 16:01:55.258  INFO 11620 --- [ntainer#0-1-C-1] c.d.t.b.l.KafkaDankalMessageListener     : 2021-01-29 16:01:55.255 执行线程: 0-1-C-1 消费消息 ==> 主题: DankalSimpleMessage(id=1, key=dankal-test-3297, content=cMFcbdTL6nimirb, sendTime=2021-01-29T16:01:25.132)
2021-01-29 16:01:55.258  INFO 11620 --- [ntainer#0-2-C-1] c.d.t.b.l.KafkaDankalMessageListener     : 2021-01-29 16:01:55.255 执行线程: 0-2-C-1 消费消息 ==> 主题: DankalSimpleMessage(id=1, key=dankal-test-7358, content=G.jLqFk]cY6]-1p, sendTime=2021-01-29T16:01:38.445)
2021-01-29 16:01:55.260  INFO 11620 --- [ntainer#0-1-C-1] c.d.t.b.l.KafkaDankalMessageListener     : 2021-01-29 16:01:55.260 执行线程: 0-1-C-1 消费消息 ==> 主题: DankalSimpleMessage(id=1, key=dankal-test-8936, content=44b[HSMoI;faqMI, sendTime=2021-01-29T16:01:32.286)

```

### 小节小述

从控制台打印我们可以发现，我们发送完请求的时间大约在 `2021-01-29 16:01:25`左右，但是呢消费者是在`2021-01-29 16:01:55`,不多不少，发送请消息和消费消息的时间间隔刚好是`30s`

## spring-kafka：批量消费消息

### spring-kafka 批量消费配置

> - listener.type : 监听者的模式
>
>   - ```java
>     public enum Type {
>     
>     			/**
>     			 * Invokes the endpoint with one ConsumerRecord at a time.
>     			 */
>     			SINGLE,
>     
>     			/**
>     			 * Invokes the endpoint with a batch of ConsumerRecords.
>     			 */
>     			BATCH
>     
>     		}
>     ```
>
>   - `spring.kafka.listener.type`
>
> - consumer.max-poll-records:  单次调用poll()时返回的最大记录数
>
>   - `spring.kafka.consumer.max-poll-records`
>
> - consumer.fetch-min-size:  服务器对一个取数请求应该返回的最小数据量
>
>   - `spring.kafka.consumer.fetch-min-size`
>
> - consumer.fetch-max-wait：如果没有足够的数据来立即满足 `fetch-min-size`给出的要求，服务器在回应fetch请求之前的最大阻塞时间。
>
>   - `spring.kafka.consumer.fetch-max-wait`

### etc.

**测试用例通过测试 `max-poll-records`来控制 批量消费的最大消费量**

```yaml
server:
  port: 9001

... ...

spring:
  kafka:
      ... ...   
      ## 一下配置用于测试 spring-kafka 的批量消费
      fetch-min-size: 16384
      max-poll-records: 3
      fetch-max-wait: 500
... ...      
```

### controller

```java
    @GetMapping("/string/batch/{count}")
    public String testBatchSendMessage(@PathVariable("count") Integer batchCount) {
        if (batchCount <= 0) {
            batchCount = 10;
        }
        CyclicBarrier cyclicBarrier = new CyclicBarrier(batchCount);
        for (int i = 0; i < batchCount; i++) {
            THREAD_POOL_EXECUTOR.submit(() -> {
                final String topic = Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE;
                final String key = "dankal-test";
                final String message = RandomUtil.randomString(20);
                try {
                    // 保证一起发送
                    cyclicBarrier.await();
                    kafkaStringMessageService.sendStringMessage(topic, key, message);
                } catch (Exception e) {
                    //ignored
                }
            });
        }
        return "success";
    }
```

### service

```java
    @Override
    public boolean sendStringMessage(String topic, String key, String message) {
        stringKafkaTemplate.send(topic, key, message);
        return true;
    }
```

### listener

```java
package cn.dankal.test.boot.listener;

import cn.dankal.test.boot.common.Common;
import cn.dankal.test.boot.common.Const;
import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.support.Acknowledgment;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;
import java.util.List;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-29
 */
@Slf4j
@Component
@ConditionalOnProperty(prefix = "spring.kafka",
        name = {"listener.type"}, havingValue = "batch")
public class BatchStringKafkaMessageListener {

    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_SIMPLE_STRING_CONSUMER
    )
    public void batchHandlerStringMessage(List<String> messageList, Acknowledgment acknowledgment) {
        if (log.isDebugEnabled()) {
            log.debug(
                    "{} 成功获取到 {} 条数据",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    messageList.size()
            );
        }
        for (String message : messageList) {
            if (log.isInfoEnabled()) {
                log.info(
                        "{} {} 成功消费消息: {}",
                        Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                        Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                        message);
            }
        }
        acknowledgment.acknowledge();
    }
}
```

### 执行请求

**http://localhost:9001/string/batch/7**

### 运行结果

![](https://img.mercymodest.com/public/20210129174507.png)

## spring-kafka: 消息消费重试

**官方文档传送门:   [dead-letters](https://docs.spring.io/spring-kafka/docs/2.2.8.RELEASE/reference/html/#dead-letters)**

> ​		Spring-Kafka 提供**消费重试**的机制。在消息**消费失败**的时候，Spring-Kafka 会通过**消费重试**机制，重新投递该消息给 Consumer ，让 Consumer 有机会重新消费消息，实现消费成功。
>
> ​		当然，Spring-Kafka 并不会无限重新投递消息给 Consumer 重新消费，而是在默认情况下，达到 N 次重试次数时，Consumer 还是消费失败时，该消息就会进入到**死信队列**。

### 消息重试配置示例

```java
package cn.dankal.test.boot.config;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.kafka.listener.DeadLetterPublishingRecoverer;
import org.springframework.kafka.listener.ErrorHandler;
import org.springframework.kafka.listener.SeekToCurrentErrorHandler;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-29
 */
@Configuration
public class KafkaBaseConfig {

    /**
     * retryCount
     */
    @Value("${spring.kafka.retry-count:3}")
    private Integer retryCount;


    /**
     * 配置 {@link SeekToCurrentErrorHandler} 用于消息消费重试
     *
     * @param template KafkaTemplate
     * @return ErrorHandler
     * @see DeadLetterPublishingRecoverer
     */
    @Bean
    @Primary
    public ErrorHandler kafkaErrorHandler(KafkaTemplate<Object, Object> template) {
        DeadLetterPublishingRecoverer deadLetterPublishingRecoverer = new DeadLetterPublishingRecoverer(template);
        return new SeekToCurrentErrorHandler(deadLetterPublishingRecoverer, retryCount);
    }
}
```

### controller

```java
    @GetMapping("/test/string")
    public String testSendStringMessage() {
        final String topic = Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE;
        final String key = "dankal-test-str-" + RandomUtil.randomInt(10, 500);
        final String message = RandomUtil.randomString(20);
        kafkaStringMessageService.sendStringMessage(topic, key, message);
        return "success";
    }
```

### service

```java
    @Override
    public boolean sendStringMessage(String topic, String key, String message) {
        stringKafkaTemplate.send(topic, key, message);
        return true;
    }
```

### listener

```java
package cn.dankal.test.boot.listener;

import cn.dankal.test.boot.common.Common;
import cn.dankal.test.boot.common.Const;
import lombok.extern.slf4j.Slf4j;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.support.Acknowledgment;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-29
 */
@Slf4j
@Component("test-retry")
public class SpringKafkaRetryTestListener {

    @KafkaListener(
            topics = {Const.Kafka.Topic.TOPIC_SIMPLE_STRING_MESSAGE},
            groupId = Const.Kafka.ConsumerGroupId.GROUP_ID_SIMPLE_STRING_CONSUMER
    )
    public void stringMessageListener(String message,
                                      Acknowledgment acknowledgment) {
        if (log.isInfoEnabled()) {
            log.info("{} 执行线程: {} 消费消息 ==> 消息: {}",
                    Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                    Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                    message);
        }
        throw new RuntimeException("throw an exception when consumer the message~");
        //acknowledgment.acknowledge();
    }
}
```

> Spring-Kafka 的消费重试功能，通过实现自定义的`SeekToCurrentErrorHandler` ，在 Consumer 消费消息异常的时候，进行拦截处理：
>
> - 在重试小于最大次数时，重新投递该消息给 Consumer ，让 Consumer 有机会重新消费消息，实现消费成功。
> - 在重试到达最大次数时，Consumer 还是消费失败时，该消息就会发送到死信队列。例如说，我们测试的 Topic 是 `"simple-string-message"` ，则其对应的死信队列的 Topic 就是 `"simple-string-message.DLT"` ，即在原有 Topic 加上 `.DLT` 后缀，就是其死信队列的 Topic 。

### 执行测试

> http://localhost:9001/test/string

### 执行结果

![](https://img.mercymodest.com/public/20210130095724.png)

> 我们在配置`SeekToCurrentErrorHandler` 设置的 `maxFailures` 为  3 ，即消费消息失败之后会最多只会重试 3 次.

## 小述 spring-kafka 的 顺序消费

> 我们先来一起了解下顺序消息的**顺序消息**的定义：
>
> - 普通顺序消息 ：Producer 将相关联的消息发送到相同的消息队列。
> - 完全严格顺序 ：在【普通顺序消息】的基础上，Consumer 严格顺序消费。
>
> 在上述的示例中，我们看到 Spring-Kafka 在 Consumer 消费消息时，**天然**就支持按照 Topic 下的 Partition 下的消息，**顺序消费**。即使在并发消息时，也能保证如此。
>
> 那么此时，我们只需要考虑将 Producer 将相关联的消息发送到 Topic 下的相同的 Partition 即可。如果我们了解 Producer 发送消息的分区策略的话，只要我们发送消息时，指定了消息的 key ，Producer 则会根据 key 的哈希值取模来获取到其在 Topic 下对应的 Partition 。
>
> 如果大家想深入理解，可以参阅博文:  [Kafka 发送消息分区选择策略详解](https://leokongwq.github.io/2017/02/27/mq-kafka-producer-partitioner.html)