# 蛋壳创意科技-技术测试-Kafka集成SpringCloudStream

> company: [蛋壳创意科技](https://www.dankal.cn/)
>
> code: [dankal-test-kafka](http://git.dankal.cn/mercyModest/dankal-test-kafka.git)
>
> spring-kafka-doc:  [spring-cloud-stream-2.1.x](https://github.com/spring-cloud/spring-cloud-stream/blob/2.1.x/docs/src/main/asciidoc/spring-cloud-stream.adoc)
>

## 梗概小述

[Spring Cloud Stream](https://github.com/spring-cloud/spring-cloud-stream/tree/2.1.x) 是一个用于构建基于**消息**的微服务应用框架，使用 Spring Integration与 Broker 进行连接。

> 啥是`Broker`?
>
> 一般来说，消息队列中间件都有一个 **Broker Server**（代理服务器），消息中转角色，负责存储消息、转发消息。
>
> 例如说在 RocketMQ 中，Broker 负责接收从生产者发送来的消息并存储、同时为消费者的拉取请求作准备。另外，Broker 也存储消息相关的元数据，包括消费者组、消费进度偏移和主题和队列消息等。

Spring Cloud Stream 提供了消息中间件的**统一抽象**，推出了 publish-subscribe、consumer groups、partition 这些统一的概念。

Spring Cloud Stream 内部有两个概念：**Binder** 和 **Binding**。

① Binder，跟消息中间件集成的组件，用来创建对应的 Binding。各消息中间件都有自己的 Binder 具体实现。

```java
public interface Binder<T, 
    // 消费者配置
    C extends ConsumerProperties, 
   // 生产者配置
    P extends ProducerProperties> { 
    
    // 创建消费者的 Binding
    Binding<T> bindConsumer(String name, String group, T inboundBindTarget, C consumerProperties);

    // 创建生产者的 Binding
    Binding<T> bindProducer(String name, T outboundBindTarget, P producerProperties);
    
}
```

② Binding，包括 Input Binding 和 Output Binding。Binding 在消息中间件与应用程序提供的 Provider 和 Consumer 之间提供了一个桥梁，实现了开发者只需使用应用程序的 Provider 或 Consumer 生产或消费数据即可，屏蔽了开发者与底层消息中间件的接触。

![](https://img.mercymodest.com/public/20210130110316.png)

## 测试用例依赖版本

> - spring-boot: 	2.1.8.RELEASE
> - ==spring-cloud:    Greenwich.SR5==
>   - spring-kafka:	2.2.8.RELEASE
>   - kafka-clients:	2.0.1
>
> 其实 `spring-kafka`和`kafka-clients`的版本被统一管理了,我们只需要关注 ==spring-cloud 版本即可
>
> - jdk:	1.8.0_241
> - gradle:   Gradle 6.7.1
>
> ### Kafka 服务相关
>
> - kafka server:    2.7.0  [集群：3]
> - zookeeper:       2.4.0  [集群:   3]

## spring-cloud-stream-kafka: helloWorld

### dependence: producer/consumer

```groovy
dependencies {
    // spring boot dependence
    implementation 'org.springframework.boot:spring-boot-starter-web'
    // spring cloud  dependence
    implementation 'org.springframework.cloud:spring-cloud-starter-stream-kafka'
}
```

#### 详情相关配置

> 请参阅：
>
> - 父工程配置: [parent-build.gradle](https://git.dankal.cn/mercyModest/dankal-test-kafka/blob/master/build.gradle)
> - producer:  []()
> - consumer: []()

### 搭建生产者

**dankal-test-kafka-spring-cloud-producer**

#### 生产者主配置文件

```yaml
server:
  port: 9002
spring:
  cloud:
    stream:
      # Bindings 配置项
      # org.springframework.cloud.stream.config.BinderProperties
      bindings:
        # 消息生产管道名称
        # 需要对应 @Output 注解
        # cn.dankal.test.cloud.producer.output.KafkaDankalMessageOutput
        kafka-dankal-message-output:
          # 消息目的地
          destination: cloud-kafka-dankal-message-topic
          # 消息内容格式: json
          content-type: application/json
      # spring cloud stream kafka 配置项
      kafka:
        # org.springframework.cloud.stream.binder.kafka.properties.KafkaBinderConfigurationProperties
        binder:
          # kafka  broker 地址,多地址逗号分隔
          brokers: 192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094
        bindings:
          # 消息生产管道名称
          # 需要对应 @Output 注解
          # cn.dankal.test.cloud.producer.output.KafkaDankalMessageOutput
          kafka-dankal-message-output:
            # kafka producer 配置项
            # org.springframework.cloud.stream.binder.kafka.properties.KafkaProducerProperties
            producer:
              # 是否同步发送消息
              # false: 采用异步的形式发送消息
              sync: true
```

#### 创建 Output Binding

```java
package cn.dankal.test.cloud.producer.output;

import org.springframework.cloud.stream.annotation.Output;
import org.springframework.messaging.MessageChannel;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
public interface KafkaDankalMessageOutput {

    /**
     * spring-cloud-stream-kafka: 测试 JSON 消息
     *
     * @return MessageChannel
     */
    @Output("kafka-dankal-message-output")
    MessageChannel kafkaDankalMessageOutput();
}
```

#### controller

```java
package cn.dankal.test.cloud.producer.controller;

import cn.dankal.test.cloud.producer.entity.DankalSimpleMessage;
import cn.dankal.test.cloud.producer.output.KafkaDankalMessageOutput;
import cn.hutool.core.util.RandomUtil;
import org.springframework.messaging.Message;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RestController;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
@RestController
public class SendMessageController {

    /**
     * KafkaDankalMessageOutput
     */
    private final KafkaDankalMessageOutput kafkaDankalMessageOutput;

    public SendMessageController(KafkaDankalMessageOutput kafkaDankalMessageOutput) {
        this.kafkaDankalMessageOutput = kafkaDankalMessageOutput;
    }

    @GetMapping("/json/{count}")
    public String sendAssignCountMessage(@PathVariable("count") Integer count) {
        if (count <= 0) {
            count = 10;
        }
        for (int i = 0; i < count; i++) {
            final String key = "dankal-cloud-kafka-" + RandomUtil.randomInt(100, 100000);
            final DankalSimpleMessage dankalSimpleMessage = new DankalSimpleMessage()
                    .setId(RandomUtil.randomInt(10, 1000000))
                    .setContent(RandomUtil.randomString(15))
                    .setSendTime(LocalDateTime.now())
                    .setKey(key);
            Message<DankalSimpleMessage> dankalSimpleMessageMessage = MessageBuilder.withPayload(dankalSimpleMessage).build();
            kafkaDankalMessageOutput.kafkaDankalMessageOutput()
                    .send(dankalSimpleMessageMessage);
        }
        return String.valueOf(count);
    }
}
```

#### 启动类

```java
package cn.dankal.test.cloud.producer;

import cn.dankal.test.cloud.producer.output.KafkaDankalMessageOutput;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.cloud.stream.annotation.EnableBinding;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-30
 */
@SpringBootApplication
@EnableBinding({KafkaDankalMessageOutput.class})
public class SpringCloudStreamKafkaProducerApplication {

    public static void main(String[] args) {
        SpringApplication.run(SpringCloudStreamKafkaProducerApplication.class, args);
    }
}
```

##### tips:

> 我们需要使用 `@EnableBinding`注解 声明指定开始 Binding 功能 扫描其 `@Output`注解，测试用例这里使用的是 `KafkaDankalMessageOutput`

### 搭建消费者

**dankal-test-kafka-spring-cloud-consumer**

#### 消息者主配置文件

```yaml
server:
  port: 9003
spring:
  cloud:
    stream:
      bindings:
        # 需要对应 @Input 注解
        kafka-dankal-message-input:
          # 消费来源
          destination: cloud-kafka-dankal-message-topic
          # 消息内容格式
          content-type: application/json
          # 消费者组
          group: cloud-kafka-dankal-message-consumer-group
          # 消费者配置
          consumer:
            # 开始多个线程(一个线程一个消费者)
            concurrency: 3
      kafka:
        binder:
          # kafka  broker 地址,多地址逗号分隔
          brokers: 192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094
```

#### 创建 接口 声明 Input Binding

```java
package cn.dankal.test.cloud.consumer.input;

import org.springframework.cloud.stream.annotation.Input;
import org.springframework.messaging.SubscribableChannel;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
public interface DankalMessageInput {

    /***
     *  input name
     */
    String INPUT_NAME = "kafka-dankal-message-input";

    /**
     * 订阅 Kafka 管道消息
     *
     * @return SubscribableChannel
     */
    @Input(INPUT_NAME)
    SubscribableChannel dankalMessageInputChannel();
}
```

#### 创建`@StreamListener`作为消费者消费对应的 Input Binding 

etc. DankalMessageInput#INPUT_NAME

```java
package cn.dankal.test.cloud.consumer.consumer;

import cn.dankal.test.cloud.consumer.common.Common;
import cn.dankal.test.cloud.consumer.common.Const;
import cn.dankal.test.cloud.consumer.entity.DankalSimpleMessage;
import cn.dankal.test.cloud.consumer.input.DankalMessageInput;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.annotation.StreamListener;
import org.springframework.messaging.handler.annotation.Payload;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
@Slf4j
@Component
public class DankalMessageConsumer {

    @StreamListener(DankalMessageInput.INPUT_NAME)
    public void onMessage(@Payload DankalSimpleMessage dankalSimpleMessage) {
        log.info(
                "{} {} 成功消费: {}",
                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                dankalSimpleMessage
        );
    }
}
```

#### 启动类

```java
package cn.dankal.test.cloud.consumer;

import cn.dankal.test.cloud.consumer.input.DankalMessageInput;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.cloud.stream.annotation.EnableBinding;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
@SpringBootApplication
@EnableBinding({DankalMessageInput.class})
public class SpringCloudStreamKafkaConsumerApplication {

    public static void main(String[] args) {
        SpringApplication.run(SpringCloudStreamKafkaConsumerApplication.class, args);
    }
}
```

我们同样需要使用 `@EnableBinding`开始 Binding 功能，扫描对应的 `@Input`注解，测试用例使用的是`DankalMessageInput`

### 执行测试

> http://localhost:9002/json/10

### 消费者运行结果

![](https://img.mercymodest.com/public/20210201120144.png)

#### tips:

> 在搭建消费者的时候，我们能将消费的并发数设置为 3
>
> ![](https://img.mercymodest.com/public/20210201120420.png)
>
> 我们可以根据实践情况对参数进行相应调整，从而实现我们需要的并发消费

## spring-cloud-stream-kafka: 消息重试

Spring-Kafka 提供**消费重试**的机制。在消息**消费失败**的时候，Spring-Kafka 会通过**消费重试**机制，重新投递该消息给 Consumer ，让 Consumer 有机会重新消费消息，实现消费成功。

> tips：Spring Cloud Stream Kafka 是基于 spring-kafka，而是在默认情况下，达到 N 次重试次数时，Consumer 还是消费失败时，该消息就会进入到死信队列。

当然，Spring-Kafka 并不会无限重新投递消息给 Consumer 重新消费，而是在默认情况下，达到 N 次重试次数时，Consumer 还是消费失败时，该消息就会进入到**死信队列**。

> 死信队列用于处理无法被正常消费的消息。当一条消息初次消费失败，Spring-Kafka 会自动进行消息重试；达到最大重试次数后，若消费依然失败，则表明消费者在正常情况下无法正确地消费该消息，此时，Spring-Kafka 不会立刻将消息丢弃，而是将其发送到该消费者对应的特殊队列中。
>
> Spring-Kafka 将这种正常情况下无法被消费的消息称为死信消息（Dead-Letter Message），将存储死信消息的特殊队列称为死信队列（Dead-Letter Queue）。后续，我们可以通过对死信队列中的消息进行重发，来使得消费者实例再次进行消费。

**而在 Kafka 中，消费重试和死信队列，是由 Spring-Kafka 所封装提供的**

详情可以参阅[蛋壳创意科技-技术测试-Kafka集群安装.md](https://git.dankal.cn/mercyModest/dankal-test-kafka/blob/master/蛋壳创意科技-技术测试-Kafka集群安装.md)

因为 Spring boot 版本问题，我们在使用 `spring-boot`整合kafka 时，虽然我们可以通过 `SeekToCurrentErrorHandler`设置消息重试，但是却无法配置消息重试时间间隔.（高版本的SpringBoot 支持）。幸运的是 在当前的 `spring-cloud-stream-kafka`版本不仅支持 消息重试配置而且我们可以就消息重试做更多的配置.etc 消息重试时间间隔。

### 修改消费者配置文件

```yaml
spring:
  cloud:
    stream:
      bindings:
        # 需要对应 @Input 注解
        kafka-dankal-message-input:
        ... ...
          consumer:
            # 开始多个线程(一个线程一个消费者)
            concurrency: 3
            # 重试次数 default:3
            maxAttempts: 4
            # 重试时间间隔的初始值 default: 1000 单位: millisecond
            backOffInitialInterval: 3000
            # 重试间隔递增系数: default: 2.0
            backOffMultiplier: 2.0
            # 重试时间间隔的最大值: default 10000 单位: millisecond
            backOffMaxInterval: 10000

      kafka:
       ... ...
        bindings:
          kafka-dankal-message-input:
            consumer:
              # 开启死信队列
              # default: false
              enableDlq: true
              # 死信队列名称: 默认为 error.{topicName}
              #dlqName:
```

### 消费者修改

```java
package cn.dankal.test.cloud.consumer.consumer;

import cn.dankal.test.cloud.consumer.common.Common;
import cn.dankal.test.cloud.consumer.common.Const;
import cn.dankal.test.cloud.consumer.entity.DankalSimpleMessage;
import cn.dankal.test.cloud.consumer.input.DankalMessageInput;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.annotation.StreamListener;
import org.springframework.messaging.handler.annotation.Payload;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
@Slf4j
@Component
public class DankalMessageConsumer {

    @StreamListener(DankalMessageInput.INPUT_NAME)
    public void onMessage(@Payload DankalSimpleMessage dankalSimpleMessage) {
        log.info(
                "{} {} 成功消费: {}",
                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                dankalSimpleMessage
        );

        // 测试 spring-cloud-stream-kafka 消息重试
        throw new RuntimeException("dankal test retry consumer message");
    }
}

```

### 浏览器执行

> http://localhost:9002/json/1

### 执行结果

![](https://img.mercymodest.com/public/20210203104303.png)

> 正如我们所看到的，消费失败之后，我们的消费者最大可以重试四次。
>
> 我们配置的初始递增时间为:  `backOffInitialInterval=3`,
>
> > 接受消息的时间是:  2021-02-03  10:30:10  ==(第一次)==
> >
> > 第一次重试时间:    2021-02-03  10:30:13  (+3) ==(第二次)==
>
> 我们配置的重试递增次数为: `backOffMultiplier=2.0`
>
> > 第二次重试时间:   2021-02-03  10:30:19 (+6) ==(第三次)==
>
> 我们配置的最大重试次数：`maxAttempts=4`，重试间隔最大值: `backOffMaxInterval=10000`ms
>
> > 第三次重试时间：2021-02-03  10:30:29(+10)  ==(第四次)==

## spring-cloud-stream-kafka: 异常处理

### 修改消费者

```java
package cn.dankal.test.cloud.consumer.consumer;

import cn.dankal.test.cloud.consumer.common.Common;
import cn.dankal.test.cloud.consumer.common.Const;
import cn.dankal.test.cloud.consumer.entity.DankalSimpleMessage;
import cn.dankal.test.cloud.consumer.input.DankalMessageInput;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.annotation.StreamListener;
import org.springframework.integration.annotation.ServiceActivator;
import org.springframework.integration.context.IntegrationContextUtils;
import org.springframework.messaging.Message;
import org.springframework.messaging.handler.annotation.Payload;
import org.springframework.messaging.support.ErrorMessage;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;
import java.util.Objects;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
@Slf4j
@Component
public class DankalMessageConsumer {

    @StreamListener(DankalMessageInput.INPUT_NAME)
    public void onMessage(@Payload DankalSimpleMessage dankalSimpleMessage) {
        log.info(
                "{} {} 成功消费: {}",
                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                dankalSimpleMessage
        );

        // 测试 spring-cloud-stream-kafka 消息重试
        throw new RuntimeException("dankal test retry consumer message");
    }

    /**
     * 处理指定 {@link  org.springframework.cloud.stream.annotation.Input} 的异常
     * <pre>
     *   异常 input name :  Channel(<destination>.<group>.errors)
     *   举个栗子:
     *         {@link DankalMessageInput#dankalMessageInputChannel()}
     *              destination:  cloud-kafka-dankal-message-topic
     *              group: cloud-kafka-dankal-message-consumer-group
     *         则 此 input 对应消费异常管道为:
     *         cloud-kafka-dankal-message-topic.cloud-kafka-dankal-message-consumer-group.errors
     *
     * </pre>
     */
    @ServiceActivator(inputChannel = "cloud-kafka-dankal-message-topic.cloud-kafka-dankal-message-consumer-group.errors")
    public void handlerDankalMessageInputError(ErrorMessage errorMessage) {
        log.info("current error is  handler to DankalMessageInput ");
        log.error("payload error message: {}", errorMessage.getPayload().getMessage());
        Message<?> originalMessage = errorMessage.getOriginalMessage();
        if (Objects.nonNull(originalMessage)) {
            log.error("origin payload message:{}", originalMessage.getPayload());
        }
        log.error("error headers:{}", errorMessage.getHeaders());
    }

    /**
     * input 全局异常异常处理
     */
    @StreamListener(IntegrationContextUtils.ERROR_CHANNEL_BEAN_NAME)
    public void globalInputErrorHandler(ErrorMessage errorMessage) {
        log.info("current error is  input error handler ");
        log.error("payload error message: {}", errorMessage.getPayload().getMessage());
        Message<?> originalMessage = errorMessage.getOriginalMessage();
        if (Objects.nonNull(originalMessage)) {
            log.error("origin payload message:{}", originalMessage.getPayload());
        }
        log.error("error headers:{}", errorMessage.getHeaders());
    }

}
```

> 当 消费`input` 消息出现异常时，`spring-cloud-stream-kafak`会异常打包成`ErrorMessage`然后发送到管道`<destination>.<group>.errors`
>
> ```java
>     /**
>      * 处理指定 {@link  org.springframework.cloud.stream.annotation.Input} 的异常
>      * <pre>
>      *   异常 input name :  Channel(<destination>.<group>.errors)
>      *   举个栗子:
>      *         {@link DankalMessageInput#dankalMessageInputChannel()}
>      *              destination:  cloud-kafka-dankal-message-topic
>      *              group: cloud-kafka-dankal-message-consumer-group
>      *         则 此 input 对应消费异常管道为:
>      *         cloud-kafka-dankal-message-topic.cloud-kafka-dankal-message-consumer-group.errors
>      *
>      * </pre>
>      */
> ```
>
> 上述处理方式单一 `Channel`异常处理，我们也可以监听`spring-cloud-stream-kafka`的全局异常`Channel`
>
> ```java
> 
>     /**
>      * input 全局异常异常处理
>      */
>     @StreamListener(IntegrationContextUtils.ERROR_CHANNEL_BEAN_NAME)
>     public void globalInputErrorHandler(ErrorMessage errorMessage) {
>         log.info("current error is  input error handler ");
>         log.error("payload error message: {}", errorMessage.getPayload().getMessage());
>         Message<?> originalMessage = errorMessage.getOriginalMessage();
>         if (Objects.nonNull(originalMessage)) {
>             log.error("origin payload message:{}", originalMessage.getPayload());
>         }
>         log.error("error headers:{}", errorMessage.getHeaders());
>     }
> ```
>
> 和`SpringMVC`类似，当某一`Channel`发生异常时，如果没有对应的`@ServiceActivator`，才会发送到全局异常`Channel`:
>
> ```java
>     @StreamListener(IntegrationContextUtils.ERROR_CHANNEL_BEAN_NAME)
> ```
>
> > tips:
> >
> > ==不过要注意，如果异常处理方法成功，没有重新抛出异常，会认定为该消息被**消费成功**，所以就不会发到死信队列了噢~==

## spring-cloud-stream-kafka: 消息过滤

**spring-cloud-stream-kafka 提供了**通用的 **Consumer** 级别的效率过滤器机制。我们只需要使用 `@StreamListener` 注解的 `condition` 属性，设置消息满足指定 Spring EL 表达式的情况下，才进行消费

### 举个栗子

#### 新建一个消费者

使用 `@StreamListener` 的 `condition` 属性结合SpEL实现消息过滤

```java

 @StreamListener(value = DankalMessageInput.INPUT_NAME,
            condition = "headers['messageTag']=='dankal-condition'")
  public void onMessage(@Payload DankalSimpleMessage dankalSimpleMessage){
      ... ...
  }
```

```java
package cn.dankal.test.cloud.consumer.consumer;

import cn.dankal.test.cloud.consumer.common.Common;
import cn.dankal.test.cloud.consumer.common.Const;
import cn.dankal.test.cloud.consumer.entity.DankalSimpleMessage;
import cn.dankal.test.cloud.consumer.input.DankalMessageInput;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.annotation.StreamListener;
import org.springframework.messaging.handler.annotation.Payload;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-04
 */
@Slf4j
@Component
public class ConditionDankalMessageConsumer {

    /**
     *  此消费者只消费  message header 带有: messageTag 并且其值为: dankal-condition
     * @param dankalSimpleMessage
     */
    @StreamListener(value = DankalMessageInput.INPUT_NAME,
            condition = "headers['messageTag']=='dankal-condition'")
    public void onMessage(@Payload DankalSimpleMessage dankalSimpleMessage) {
        log.info(
                "ConditionDankalMessageConsumer: {} {} 成功消费: {}",
                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                dankalSimpleMessage
        );
    }
}
```

#### controller 

```java
    @GetMapping("/condition/{count}")
    public String sendAssignCountConditionMessage(@PathVariable("count") Integer count) {
        if (count <= 0) {
            count = 10;
        }
        int setHeaderTagCount = 0;
        for (int i = 0; i < count; i++) {
            final String key = "dankal-cloud-kafka-" + RandomUtil.randomInt(100, 100000);
            final DankalSimpleMessage dankalSimpleMessage = new DankalSimpleMessage()
                    .setId(RandomUtil.randomInt(10, 1000000))
                    .setContent(RandomUtil.randomString(15))
                    .setSendTime(LocalDateTime.now())
                    .setKey(key);
            MessageBuilder<DankalSimpleMessage> messageBuilder = MessageBuilder
                    .withPayload(dankalSimpleMessage);
            if (RandomUtil.randomBoolean()) {
                // 随机添加 message  header tag
                messageBuilder.setHeader("messageTag", "dankal-condition");
                setHeaderTagCount++;
            }
            Message<DankalSimpleMessage> dankalSimpleMessageMessage = messageBuilder
                    .build();
            kafkaDankalMessageOutput.kafkaDankalMessageOutput()
                    .send(dankalSimpleMessageMessage);
        }
        log.info("共发送 {} 条消息,其中 {} 条 带有 header tag", count, setHeaderTagCount);
        return String.valueOf(count);
    }
```

#### 执行测试

> http://localhost:9002/condition/5

#### producer 控制台

![](https://img.mercymodest.com/public/20210204180251.png)

#### consumer 控制台

##### ConditionDankalMessageConsumer

![](https://img.mercymodest.com/public/20210204180410.png)

##### DankalMessageConsumer

![](https://img.mercymodest.com/public/20210204180525.png)

## spring-cloud-stream-kafka: 消息的手动提交

**spring-cloud-stream-kafka 的提交机制**

spring-cloud-stream-kafka 在 spring-kafka 上进一步封装，在`spring.cloud.stream.kafka.bindings..<bindingName>.consumer` 下提供了两个配置项：

- `auto-commit-offset` 配置项，是否自动提交消费进度，默认为 `true` 自动提交。
- `ack-each-record` 配置项，是否每一条消息都进行提交消费进度，默认为 `false` 在每一批消费完成后一起提交。

我们进行下整理，将 Spring Cloud Stream Kafka 这两个配置项，和 Spring-Kafka 的 AckMode 对应上，如下表格：

| AckMode                 | `auto-commit-offset` | `ack-each-record` |
| :---------------------- | :------------------- | :---------------- |
| 自动 `RECORD`           | `true`               | `false`           |
| 自动 `BATCH`            | `true`               | `true`            |
| 手动 `MANUAL`           | `false`              | `false`           |
| 手动 `MANUAL_IMMEDIATE` | `false`              | `true`            |

因此，**默认什么都不配置的情况下，也使用 Spring-Kafka 的 BATCH 模式：每一次消息被消费完成后，在下次拉取消息之前，自动提交**。

### 修改配置文件

```yaml
server:
  port: 9003
spring:
  cloud:
    stream:
      ... ...
      kafka:
        binder:
          # kafka  broker 地址,多地址逗号分隔
          brokers: 192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094
        bindings:
          kafka-dankal-message-input:
            consumer:
              # 开启死信队列
              # default: false
              enableDlq: true
              # 死信队列名称: 默认为 error.{topicName}
              #dlqName:
              # 是否取消自动提交
              # default: true
              autoCommitOffset: false
              # 是否每消费一条消息 提交一次偏移量
              # default: false 批量提交
              # 即 全部 消费完 单次 poll 的消息后,一次性提交偏移量
              ackEachRecord: true
```

### 修改消费者

```java
@Slf4j
@Component
public class DankalMessageConsumer {

    @StreamListener(DankalMessageInput.INPUT_NAME)
    public void onMessage(
            @Payload DankalSimpleMessage dankalSimpleMessage,
            @Header(KafkaHeaders.ACKNOWLEDGMENT) Acknowledgment acknowledgment) {
        log.info(
                "DankalMessageConsumer: {} {} 成功消费: {}",
                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                dankalSimpleMessage
        );
        if (dankalSimpleMessage.getId() % 2 == 0) {
            acknowledgment.acknowledge();
            log.info("{} 消费确认", dankalSimpleMessage.getId());
        }

        // 测试 spring-cloud-stream-kafka 消息重试
        //throw new RuntimeException("dankal test retry consumer message");
    }
}    
```

### controller

```java
    @GetMapping("/json/{count}")
    public String sendAssignCountMessage(@PathVariable("count") Integer count) {
        if (count <= 0) {
            count = 10;
        }
        for (int i = 0; i < count; i++) {
            final String key = "dankal-cloud-kafka-" + RandomUtil.randomInt(100, 100000);
            final DankalSimpleMessage dankalSimpleMessage = new DankalSimpleMessage()
                    .setId(RandomUtil.randomInt(10, 1000000))
                    .setContent(RandomUtil.randomString(15))
                    .setSendTime(LocalDateTime.now())
                    .setKey(key);
            Message<DankalSimpleMessage> dankalSimpleMessageMessage = MessageBuilder.withPayload(dankalSimpleMessage).build();
            kafkaDankalMessageOutput.kafkaDankalMessageOutput()
                    .send(dankalSimpleMessageMessage);
        }
        return String.valueOf(count);
    }
```

### 请求执行

> http://localhost:9002/json/5

### consumer 截图

![](https://img.mercymodest.com/public/20210204183027.png)

> 全部消息了发送 5 条消息，但是只有 1条消息提交了偏移量

![](https://img.mercymodest.com/public/20210204183153.png)